#if defined _bossengine_egg_included
    #endinput
#endif
#define _bossengine_egg_included

#include <sourcemod>

#define INVALID_EGG view_as<Egg>(-1)

static Handle s_fwdEggCreated;
static Handle s_fwdEggDestroyed;

static StringMap s_trieFastLookup;
static int s_eggCounter;
static ArrayList s_listEggs;
static int s_lastEggID;
static StringMap s_lastEggData;

static bool s_bEggNativesActive;

methodmap Egg
{
    public static void Link()
    {
        CreateNative("BossEngine_CreateEgg", NativeCall_CreateEgg);
        CreateNative("BossEngine_IsValidEgg", NativeCall_IsValidEgg);
        CreateNative("BossEngine_FindEggByName", NativeCall_FindEggByName);
        CreateNative("Egg.Destroy", NativeCall_Egg_Destroy);
        CreateNative("Egg.GetName", NativeCall_Egg_GetName);
        CreateNative("Egg.HasAttribute", NativeCall_Egg_HasAttribute);
        CreateNative("Egg.GetAttribute", NativeCall_Egg_GetAttribute);
        CreateNative("Egg.SetAttribute", NativeCall_Egg_SetAttribute);
        CreateNative("Egg.ClearAttribute", NativeCall_Egg_ClearAttribute);
        CreateNative("Egg.DumpAttributes", NativeCall_Egg_DumpAttributes);
        CreateNative("Egg.LoadConfig", NativeCall_Egg_LoadConfig);
        
        s_bEggNativesActive = true;
        return;
    }
    
    public static void Init()
    {
        s_fwdEggCreated = CreateGlobalForward("BossEngine_EggCreated", ET_Ignore, Param_Cell, Param_String);
        s_fwdEggDestroyed = CreateGlobalForward("BossEngine_EggDestroyed", ET_Ignore, Param_Cell, Param_String);
        
        s_trieFastLookup = new StringMap();
        s_eggCounter = 0;
        s_listEggs = new ArrayList(2);
        s_lastEggID = 0;
        s_lastEggData = null;
        return;
    }
    
    public Egg(Handle creator, const char[] name, Function callback)
    {
        StringMap trieData = new StringMap();
        if (trieData == null)
        {
            return INVALID_EGG;
        }
        
        Handle fwdHeartbeat = CreateForward(ET_Ignore);
        AddToForward(fwdHeartbeat, creator, callback);
        
        int cell = s_listEggs.Push(++s_eggCounter);
        s_listEggs.Set(cell, trieData, 1);
        
        trieData.SetValue("creator", creator);
        trieData.SetString("name", name);
        trieData.SetValue("heartbeat", fwdHeartbeat);
        trieData.SetValue("attributes", new StringMap());
        trieData.SetValue("configs", new StringMap());
        
        s_trieFastLookup.SetValue(name, s_eggCounter);
        
        s_bEggNativesActive = false;
        {
            Call_StartForward(s_fwdEggCreated);
            Call_PushCell(s_eggCounter);
            Call_PushString(name);
            Call_Finish();
        }
        s_bEggNativesActive = true;
        
        return view_as<Egg>(s_eggCounter);
    }
    
    property int Index
    {
        public get()
        {
            return view_as<int>(this);
        }
    }
    
    public void Refresh()
    {
        if (this.Index != s_lastEggID)
        {
            int cell = s_listEggs.FindValue(this);
            if (cell == -1)
            {
                ThrowError("Egg '%i' not found", this.Index);
            }
            
            s_lastEggID = this.Index;
            s_lastEggData = s_listEggs.Get(cell, 1);
        }
        return;
    }
    
    property Handle Creator
    {
        public get()
        {
            Handle hCreator;
            this.Refresh();
            s_lastEggData.GetValue("creator", hCreator);
            return hCreator;
        }
    }
    
    property Handle Heartbeat
    {
        public get()
        {
            Handle hHeartbeat;
            this.Refresh();
            s_lastEggData.GetValue("heartbeat", hHeartbeat);
            return hHeartbeat;
        }
    }
    
    property bool IsAlive
    {
        public get()
        {
            this.Refresh();
            return (GetForwardFunctionCount(this.Heartbeat) != 0);
        }
    }
    
    property StringMap Attributes
    {
        public get()
        {
            StringMap output;
            this.Refresh();
            s_lastEggData.GetValue("attributes", output);
            return output;
        }
    }
    
    property StringMap Configs
    {
        public get()
        {
            StringMap output;
            this.Refresh();
            s_lastEggData.GetValue("configs", output);
            return output;
        }
    }
    
    public void GetName(char[] szBuffer, int maxlength)
    {
        this.Refresh();
        s_lastEggData.GetString("name", szBuffer, maxlength);
        return;
    }
    
    public bool HasAttribute(const char[] name)
    {
        StringMap data = this.Attributes;
        if (data == null)
        {
            return false;
        }
        char szBuffer[3];
        return data.GetString(name, szBuffer, sizeof(szBuffer));
    }
    
    public bool GetAttribute(const char[] name, char[] buffer, int maxlength)
    {
        StringMap data = this.Attributes;
        if (data == null)
        {
            return false;
        }
        return data.GetString(name, buffer, maxlength);
    }
    
    public void SetAttribute(const char[] key, const char[] value)
    {
        if (!this.IsAlive)
        {
            return;
        }
        
        StringMap data = this.Attributes;
        if (data != null)
        {
            data.SetString(key, value);
        }
        return;
    }
    
    public void ClearAttribute(const char[] key)
    {
        if (!this.IsAlive)
        {
            return;
        }
        StringMap data = this.Attributes;
        if (data != null)
        {
            data.Remove(key);
        }
        return;
    }
    
    public void Destroy()
    {
        this.Refresh();
        
        char szName[32];
        s_lastEggData.GetString("name", szName, sizeof(szName));
        
        Handle buffer;
        s_lastEggData.GetValue("heartbeat", buffer);
        delete buffer;
        
        s_lastEggData.GetValue("attributes", buffer);
        delete buffer;
        
        s_lastEggData.GetValue("configs", buffer);
        delete buffer;
        
        delete s_lastEggData;
        s_lastEggData = null;
        
        int cell = s_listEggs.FindValue(this);
        if (cell != -1)
        {
            s_listEggs.Erase(cell);
        }
        
        s_bEggNativesActive = false;
        {
            Call_StartForward(s_fwdEggDestroyed);
            Call_PushCell(s_lastEggID);
            Call_PushString(szName);
            Call_Finish();
        }
        s_bEggNativesActive = true;
        
        s_lastEggID = 0;
        return;
    }
    
    public bool LoadConfig(const char[] filename)
    {
        char szConfigPath[PLATFORM_MAX_PATH];
        BuildPath(Path_SM, szConfigPath, PLATFORM_MAX_PATH, "configs/bossengine/%s.cfg", filename);
        if (!FileExists(szConfigPath))
        {
            return false;
        }
        
        int utcNew = GetFileTime(filename, FileTime_LastChange);
        int utcOld;
        if (this.Configs.GetValue(filename, utcOld))
        {
            if (utcNew == utcOld)
            {
                return false;
            }
        }
        this.Configs.SetValue(filename, utcNew);
        
        KeyValues kvBuffer = new KeyValues("");
        if (kvBuffer == null)
        {
            LogError("KeyValue object creation failure");
            return false;
        }
        
        kvBuffer.ImportFromFile(szConfigPath);
        kvBuffer.Rewind();
        
        if (kvBuffer.JumpToKey("configs"))
        {
            kvBuffer.GoBack();
        }
        
        if (kvBuffer.JumpToKey("add-attributes"))
        {
            if (kvBuffer.GotoFirstSubKey())
            {
                do
                {
                    char szName[64];
                    kvBuffer.GetString("name", szName, sizeof(szName));
                    
                    if (szName[0] == '\0')
                    {
                        continue;
                    }
                    
                    char szValue[128];
                    kvBuffer.GetString("value", szValue, sizeof(szValue));
                    
                    this.SetAttribute(szName, szValue);
                }
                while (kvBuffer.GotoNextKey());
                kvBuffer.GoBack();
            }
            kvBuffer.GoBack();
        }
        
        if (kvBuffer.JumpToKey("del-attributes"))
        {
            if (kvBuffer.GotoFirstSubKey())
            {
                do
                {
                    char szName[64];
                    kvBuffer.GetString("name", szName, sizeof(szName));
                    
                    if (szName[0] == '\0')
                    {
                        continue;
                    }
                    
                    this.ClearAttribute(szName);
                }
                while (kvBuffer.GotoNextKey());
                kvBuffer.GoBack();
            }
            kvBuffer.GoBack();
        }
        
        delete kvBuffer;
        return true;
    }
    
    public void DumpAttributes()
    {
        if (!this.IsAlive)
        {
            return;
        }
        StringMap data = this.Attributes;
        if (data != null)
        {
            data.Clear();
        }
        return;
    }
    
    public static void GCThink()
    {
        int count = GetArraySize(s_listEggs);
        for (int cell = count - 1; cell >= 0; cell--)
        {
            Egg item = s_listEggs.Get(cell, 0);
            if (!item.IsAlive)
            {
                item.Destroy();
                count--;
            }
        }
        return;
    }
}

stock bool IsValidEgg(Egg data)
{
    return (s_listEggs.FindValue(data) != -1);
}

public int NativeCall_CreateEgg(Handle plugin, int numParams)
{
    char szName[128];
    GetNativeString(1, szName, sizeof(szName));
    
    char szBuffer[3];
    if (s_trieFastLookup.GetString(szName, szBuffer, sizeof(szBuffer)))
    {
        //return view_as<int>(INVALID_EGG);
        return -1;      // these two lines are infact identical, as INVALID_EGG === view_as<Egg>(-1)
    }
    
    Function callback = GetFunctionByName(plugin, "BossEngine_Heartbeat");
    if (callback == INVALID_FUNCTION)
    {
        ThrowNativeError(SP_ERROR_INVALID_ADDRESS, "Could not locate heartbeat callback ('BossEngine_Heartbeat')");
    }
    
    if (!s_bEggNativesActive)
    {
        return -1;
    }
    
    Egg output = Egg(plugin, szName, callback);
    return view_as<int>(output);
}

public int NativeCall_IsValidEgg(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    return IsValidEgg(item);
}

public int NativeCall_FindEggByName(Handle plugin, int numParams)
{
    int length = 0;
    GetNativeStringLength(1, length);
    char[] szName = new char[length+1];
    GetNativeString(1, szName, length+1);
    
    Egg item = INVALID_EGG;
    if (s_trieFastLookup.GetValue(szName, item))
    {
        if (IsValidEgg(item))
        {
            return item.Index;
        }
    }
    return -1;
}

public int NativeCall_Egg_Destroy(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid egg reference");
    }
    else if (item.Creator != plugin)
    {
        ThrowNativeError(SP_ERROR_MEMACCESS, "Only an egg creator may destroy an egg");
    }
    
    if (!s_bEggNativesActive)
    {
        return;
    }
    
    item.Destroy();
    return;
}

public int NativeCall_Egg_HasAttribute(Handle plugin, numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid egg index");
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    return item.HasAttribute(szName);
}

public int NativeCall_Egg_GetAttribute(Handle plugin, numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid egg index");
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    length = GetNativeCell(4);
    char[] szValue = new char[length+1];
    
    if (item.GetAttribute(szName, szValue, length+1))
    {
        SetNativeString(3, szValue, length+1);
        return 1;
    }
    return 0;
}

public int NativeCall_Egg_SetAttribute(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid egg reference");
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    GetNativeStringLength(3, length);
    char[] szValue = new char[length+1];
    GetNativeString(3, szValue, length+1);
    
    item.SetAttribute(szName, szValue);
    return;
}

public int NativeCall_Egg_ClearAttribute(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid egg reference");
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    item.ClearAttribute(szName);
    return;
}

public int NativeCall_Egg_DumpAttributes(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid egg reference");
    }
    
    item.DumpAttributes();
    return;
}

public int NativeCall_Egg_LoadConfig(Handle plugin, int numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid egg reference");
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szFilename = new char[length+1];
    GetNativeString(2, szFilename, length+1);
    return view_as<int>(item.LoadConfig(szFilename));
}

public int NativeCall_Egg_GetName(Handle plugin, numParams)
{
    Egg item = GetNativeCell(1);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid egg index");
    }
    
    int length = GetNativeCell(3);
    char[] szName = new char[length+1];
    item.GetName(szName, length+1);
    
    SetNativeString(2, szName, length);
}
