#if defined _bossengine_monster_included
    #endinput
#endif
#define _bossengine_monster_included

#define INVALID_MONSTER view_as<Monster>(-1)

static Handle s_fwdMonsterCreated;
static Handle s_fwdMonsterDestroyed;

static int s_monsterCounter;

static ArrayList s_listMonsters;
static int s_lastMonsterID;
static int s_lastMonsterClient;
static StringMap s_lastMonsterData;

static bool s_bMonsterNativesActive;

methodmap Monster
{
    public static void Link()
    {
        CreateNative("BossEngine_CreateMonster", NativeCall_CreateMonster);
        CreateNative("BossEngine_IsValidMonster", NativeCall_IsValidMonster);
        CreateNative("BossEngine_GetClientMonster", NativeCall_GetClientMonster);
        CreateNative("Monster.Client.get", NativeCall_Monster_Client_get);
        CreateNative("Monster.EggID.get", NativeCall_Monster_EggID_get);
        CreateNative("Monster.Destroy", NativeCall_Monster_Destroy);
        CreateNative("Monster.GetName", NativeCall_Monster_GetName);
        CreateNative("Monster.HasProperty", NativeCall_Monster_HasProperty);
        CreateNative("Monster.GetProperty", NativeCall_Monster_GetProperty);
        CreateNative("Monster.SetProperty", NativeCall_Monster_SetProperty);
        CreateNative("Monster.ClearProperty", NativeCall_Monster_ClearProperty);
        CreateNative("Monster.HasAttribute", NativeCall_Monster_HasAttribute);
        CreateNative("Monster.GetAttribute", NativeCall_Monster_GetAttribute);
        return;
    }
    
    public static void Init()
    {
        s_fwdMonsterCreated = CreateGlobalForward("BossEngine_MonsterCreated", ET_Ignore, Param_Cell, Param_String, Param_Cell);
        s_fwdMonsterDestroyed = CreateGlobalForward("BossEngine_MonsterDestroyed", ET_Ignore, Param_Cell, Param_String, Param_Cell);
        
        s_monsterCounter = 0;
        s_listMonsters = new ArrayList(3);
        s_lastMonsterID = -1;
        s_lastMonsterClient = -1;
        s_lastMonsterData = null;
        
        s_bMonsterNativesActive = true;
        return;
    }
    
    public Monster(int client, Egg item, bool bBroadcast = true)
    {
        if (!IsValidClient(client) || !IsValidEgg(item))
        {
            return INVALID_MONSTER;
        }
        else if (GetClientMonster(client) != INVALID_MONSTER)
        {
            // only one monster per client :3
            return INVALID_MONSTER;
        }
        
        StringMap trieData = new StringMap();
        if (trieData == null)
        {
            return INVALID_MONSTER;
        }
        
        int cell = s_listMonsters.Push(++s_monsterCounter);
        s_listMonsters.Set(cell, trieData, 1);
        s_listMonsters.Set(cell, client, 2);
        
        char szName[64];
        item.GetName(szName, sizeof(szName));
        trieData.SetString("name", szName);
        
        trieData.SetValue("egg", item.Index);
        
        Handle buffer;
        
        buffer = item.Creator
        trieData.SetValue("creator", buffer);
        
        buffer = CloneHandle(item.Heartbeat);
        trieData.SetValue("heartbeat", buffer);
        
        buffer = CloneAttributes(item.Attributes);
        trieData.SetValue("attributes", buffer);
        
        buffer = new StringMap();
        trieData.SetValue("properties", buffer);
        
        s_bMonsterNativesActive = false;
        {
            Call_StartForward(s_fwdMonsterCreated);
            Call_PushCell(s_monsterCounter);
            Call_PushString(szName);
            Call_PushCell(client);
            Call_Finish();
        }
        s_bMonsterNativesActive = true;
        
        return view_as<Monster>(s_monsterCounter);
    }
    
    property int Index
    {
        public get()
        {
            return view_as<int>(this);
        }
    }
    
    
    public void Refresh()
    {
        if (this.Index != s_lastMonsterID)
        {
            int cell = s_listMonsters.FindValue(this);
            if (cell == -1)
            {
                ThrowError("Monster '%i' not found", this.Index);
            }
            
            s_lastMonsterID = this.Index;
            s_lastMonsterData = s_listMonsters.Get(cell, 1);
            s_lastMonsterClient = s_listMonsters.Get(cell, 2);
        }
        return;
    }
    
    public void GetName(char[] szBuffer, int maxlength)
    {
        this.Refresh();
        s_lastMonsterData.GetString("name", szBuffer, maxlength);
        return;
    }
    
    property Handle Heartbeat
    {
        public get()
        {
            Handle buffer;
            this.Refresh();
            if (s_lastMonsterData.GetValue("heartbeat", buffer))
            {
                return buffer;
            }
            else
            {
                return null;
            }
        }
    }
    
    property StringMap Props
    {
        public get()
        {
            StringMap buffer;
            this.Refresh();
            if (s_lastMonsterData.GetValue("properties", buffer))
            {
                return buffer;
            }
            else
            {
                return null;
            }
        }
    }
    
    property StringMap Attributes
    {
        public get()
        {
            StringMap buffer;
            this.Refresh();
            if (s_lastMonsterData.GetValue("attributes", buffer))
            {
                return buffer;
            }
            else
            {
                return null;
            }
        }
    }
    
    property bool IsAlive
    {
        public get()
        {
            this.Refresh();
            if (this.Heartbeat == null)
            {
                return false;
            }
            else
            {
                return (GetForwardFunctionCount(this.Heartbeat) != 0);
            }
        }
    }
    
    property int Client
    {
        public get()
        {
            this.Refresh();
            return s_lastMonsterClient;
        }
    }
    
    property int EggID
    {
        public get()
        {
            int buffer;
            this.Refresh();
            s_lastMonsterData.GetValue("egg", buffer);
            return buffer;
        }
    }
    
    public bool HasAttribute(const char[] name)
    {
        StringMap data = this.Attributes;
        if (data == null)
        {
            return false;
        }
        char szBuffer[3];
        return data.GetString(name, szBuffer, sizeof(szBuffer));
    }
    
    public bool GetAttribute(const char[] name, char[] buffer, int maxlength)
    {
        StringMap data = this.Attributes;
        if (data == null)
        {
            return false;
        }
        
        if (data.GetString(name, buffer, maxlength))
        {
            if (buffer[0] != '\0')
            {
                return true;
            }
        }
        return false;
    }
    
    public bool HasProperty(const char[] name)
    {
        if (this.HasAttribute(name))
        {
            return true;
        }
        
        StringMap data = this.Props;
        if (data == null)
        {
            return false;
        }
        char szBuffer[3];
        return data.GetString(name, szBuffer, sizeof(szBuffer));
    }
    
    public bool GetProperty(const char[] name, char[] buffer, int maxlength)
    {
        if (this.GetAttribute(name, buffer, maxlength))
        {
            return true;
        }
        
        StringMap data = this.Props;
        if (data == null)
        {
            return false;
        }
        
        if (data.GetString(name, buffer, maxlength))
        {
            if (buffer[0] != '\0')
            {
                return true;
            }
        }
        return false;
    }
    
    public bool SetProperty(const char[] key, const char[] value)
    {
        if (!this.IsAlive)
        {
            return false;
        }
        
        if (this.HasAttribute(key))
        {
            return false;
        }
        
        StringMap data = this.Props;
        if (data == null)
        {
            return false;
        }
        
        data.SetString(key, value);
        return true
    }
    
    public bool ClearProperty(const char[] key)
    {
        if (!this.IsAlive)
        {
            return false;
        }
        
        if (this.HasAttribute(key))
        {
            return false;
        }
        
        StringMap data = this.Props;
        if (data == null)
        {
            return false;
        }
        
        data.Remove(key);
        return true;
    }
    
    public void Destroy()
    {
        this.Refresh();
        
        char szName[64];
        s_lastMonsterData.GetString("name", szName, sizeof(szName));
        
        Handle buffer;
        
        s_lastMonsterData.GetValue("properties", buffer);
        s_lastMonsterData.Remove("properties");
        delete buffer;
        
        s_bMonsterNativesActive = false;
        {
            Call_StartForward(s_fwdMonsterDestroyed);
            Call_PushCell(this);
            Call_PushString(szName);
            Call_PushCell(s_lastMonsterClient);
            Call_Finish();
        }
        s_bMonsterNativesActive = true;
        
        s_lastMonsterData.GetValue("heartbeat", buffer);
        s_lastMonsterData.Remove("heartbeat");
        delete buffer;
        
        s_lastMonsterData.GetValue("attributes", buffer);
        s_lastMonsterData.Remove("attributes");
        delete buffer;
        
        delete s_lastMonsterData;
        s_lastMonsterData = null;
        
        int cell = s_listMonsters.FindValue(this);
        if (cell != -1)
        {
            s_listMonsters.Erase(cell);
        }
        
        s_lastMonsterID = -1;
        s_lastMonsterData = null;
        s_lastMonsterClient = -1;
        
        return;
    }
    
    public static void GCThink()
    {
        for (int player = 0; player <= MAXPLAYERS; player++)
        {
            Monster thing = GetClientMonster(player);
            if (thing != INVALID_MONSTER)
            {
                if (!thing.IsAlive)
                {
                    thing.Destroy();
                }
            }
        }
        return;
    }
}

stock bool IsValidMonster(Monster thing)
{
    return (s_listMonsters.FindValue(thing, 0) != -1);
}

stock Monster GetClientMonster(int client)
{
    int cell = s_listMonsters.FindValue(client, 2);
    if (cell != -1)
    {
        return s_listMonsters.Get(cell, 0);
    }
    return INVALID_MONSTER;
}

public int NativeCall_CreateMonster(Handle plugin, numParams)
{
    int target = GetNativeCell(1);
    if (!IsValidClient(target))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid client index");
    }
    
    Egg item = GetNativeCell(2);
    if (!IsValidEgg(item))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid egg index");
    }
    
    if (!s_bMonsterNativesActive)
    {
        return -1;
    }
    
    Monster output = Monster(target, item);
    return view_as<int>(output);
}

public int NativeCall_IsValidMonster(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    return IsValidMonster(thing);
}

public int NativeCall_GetClientMonster(Handle plugin, numParams)
{
    int client = GetNativeCell(1);
    if (!IsValidClient(client))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid client index");
        return -1;
    }
    return view_as<int>(GetClientMonster(client));
}

public int NativeCall_Monster_Client_get(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return -1;
    }
    return thing.Client;
}

public int NativeCall_Monster_EggID_get(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return -1;
    }
    return thing.EggID;
}

public int NativeCall_Monster_Destroy(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    if (!s_bMonsterNativesActive)
    {
        return -1;
    }
    
    if (thing.IsAlive)
    {
        thing.Destroy();
    }
    return 0;
}

public int NativeCall_Monster_GetName(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    int length = GetNativeCell(3);
    char[] szName = new char[length+1];
    thing.GetName(szName, length+1);
    
    SetNativeString(2, szName, length);
    return 0;
}

public int NativeCall_Monster_HasProperty(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    return thing.HasProperty(szName);
}

public int NativeCall_Monster_GetProperty(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    length = GetNativeCell(4);
    char[] szValue = new char[length+1];
    
    if (thing.GetProperty(szName, szValue, length+1))
    {
        SetNativeString(3, szValue, length+1);
        return 1;
    }
    return 0;
}

public int NativeCall_Monster_SetProperty(Handle plugin, int numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid monster reference");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    GetNativeStringLength(3, length);
    char[] szValue = new char[length+1];
    GetNativeString(3, szValue, length+1);
    
    return thing.SetProperty(szName, szValue);
}

public int NativeCall_Monster_ClearProperty(Handle plugin, int numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_INDEX, "Invalid monster reference");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    return thing.ClearProperty(szName);
}

public int NativeCall_Monster_HasAttribute(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    return thing.HasAttribute(szName);
}

public int NativeCall_Monster_GetAttribute(Handle plugin, numParams)
{
    Monster thing = GetNativeCell(1);
    if (!IsValidMonster(thing))
    {
        ThrowNativeError(SP_ERROR_PARAM, "Invalid monster index");
        return 0;
    }
    
    int length = 0;
    GetNativeStringLength(2, length);
    char[] szName = new char[length+1];
    GetNativeString(2, szName, length+1);
    
    length = GetNativeCell(4);
    char[] szValue = new char[length+1];
    
    if (thing.GetAttribute(szName, szValue, length+1))
    {
        SetNativeString(3, szValue, length+1);
        return 1;
    }
    return 0;
}


//--------//
// STOCKS //
//--------//
stock StringMap CloneAttributes(StringMap input)
{
    if (input == null)
    {
        return null;
    }
    
    StringMap output = new StringMap();
    StringMapSnapshot data = input.Snapshot();
    int keyCount = data.Length;
    
    for (int key = 0; key < keyCount; key++)
    {
        int keyLength = data.KeyBufferSize(key) + 1;
        char[] szKey = new char[keyLength];
        data.GetKey(key, szKey, keyLength);
        
        char szValue[256];
        input.GetString(szKey, szValue, sizeof(szValue));
        output.SetString(szKey, szValue);
    }
    delete data;
    return output;
}

stock bool IsValidClient(int client)
{
    if ((client < 1) || (client > MaxClients))
    {
        return false;
    }
    return IsClientInGame(client);
}
